/**
 * goIRC (C)opyright 2007 Sebastian Kasparek
 *
 * application.cpp - commandline and settings reader and controller in mvc like architecture
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "application.h"
#include "messages.h"
#include "mainwindow.h"
#include "socket.h"
#include "ircparser.h"

// FIXME: take care of pingtimeout

class App::MyPrivate
{
	public:
		MyPrivate();
		os::String InputEncoding( os::String cInput );
		os::String OutputEncoding( os::String cOutput );
		class IrcConnection
		{
			public:
				os::String m_cNick;
				os::String m_cServer;
				int m_nPort;
				os::String m_cPassword;
				SocketThread* m_pcSocketThread;
				os::Messenger m_cMessengerSocket;
				std::map<os::String,os::String> m_cChannels;
				os::String m_cChannel;
				os::String m_cDefaultEncoding;
				iconv_t m_hIconvIn;
				iconv_t m_hIconvOut;
		};
		IrcConnection* m_pcConn;
		IrcParser* m_pcIrcParser;
		MainWindow* m_pcMainWindow;
		os::Messenger m_cMessengerWindowMain;
		os::Settings m_cSettings;
};

App::MyPrivate::MyPrivate()
{
	m_pcIrcParser = new IrcParser();
	m_pcConn = new IrcConnection();
	m_pcConn->m_cNick = "";
	m_pcConn->m_cServer = "";
	m_pcConn->m_nPort = 0;
	m_pcConn->m_cPassword = "";
	m_pcConn->m_cDefaultEncoding = "";
	m_pcConn->m_hIconvIn = (iconv_t)-1;
	m_pcConn->m_hIconvIn = (iconv_t)-1;
}

os::String App::MyPrivate::InputEncoding( os::String cInput )
{
	// do nothing if no valid Encoding is set
	if( m_pcConn->m_hIconvIn == (iconv_t)-1 ) return( cInput );

	char anInBuf[4096];
	char* pInBuf = &anInBuf[0];
	size_t nInBufSize = cInput.Length();
	for( unsigned int i = 0; i < nInBufSize; i++ ) anInBuf[i] = cInput[i];
	
	char anOutBuf[4096];
	char* pOutBuf = &anOutBuf[0];
	size_t nOutBufSize = 4096;
	
	iconv( m_pcConn->m_hIconvIn, &pInBuf, &nInBufSize, &pOutBuf, &nOutBufSize );
	anOutBuf[4096-nOutBufSize] = 0;
	
	return( os::String( anOutBuf ) );
}

os::String App::MyPrivate::OutputEncoding( os::String cOutput )
{
	// do nothing if no valid Encoding is set
	if( m_pcConn->m_hIconvOut == (iconv_t)-1 ) return( cOutput );

	char anInBuf[4096];
	char* pInBuf = &anInBuf[0];
	size_t nInBufSize = cOutput.Length();
	for( unsigned int i = 0; i < nInBufSize; i++ ) anInBuf[i] = cOutput[i];
	
	char anOutBuf[4096];
	char* pOutBuf = &anOutBuf[0];
	size_t nOutBufSize = 4096;
	
	iconv( m_pcConn->m_hIconvOut, &pInBuf, &nInBufSize, &pOutBuf, &nOutBufSize );
	anOutBuf[4096-nOutBufSize] = 0;
	
	return( os::String( anOutBuf ) );
}

App::App( os::OptionParser* pcOpts ) : os::Application( "application/x-vnd.kasbah-goIRC" )
{
	my = new MyPrivate();

	{  // read the settings from commandline, or if not given, from settings file
		os::OptionParser::option const* pcOption;

		os::Settings cSettings;
		cSettings.Load();

		// FIXME: in contrast to the documentation the OptionParser does not recognize values for options in the shortform
		// "Both short-options and long-options can have additional arguments. If a short-option take an additional argument
		//  it the next option will be used as the argument value."
		// e.g. -n kasbah does NOT work while --nick=kasbah does work
		if( ( pcOption = pcOpts->FindOption( 'n' ) ) != 0 && pcOption->has_value )
			my->m_pcConn->m_cNick = pcOption->value.const_str();
		else
			my->m_pcConn->m_cNick = cSettings.GetString( "nick" );

		if( ( pcOption = pcOpts->FindOption( 's' ) ) != 0 && pcOption->has_value )
			my->m_pcConn->m_cServer = pcOption->value.const_str();
		else
			my->m_pcConn->m_cServer = cSettings.GetString( "server" );

		if( ( pcOption = pcOpts->FindOption( 'p' ) ) != 0 && pcOption->has_value )
			my->m_pcConn->m_nPort = atoi( pcOption->value.c_str() );
		else
			my->m_pcConn->m_nPort = cSettings.GetInt16( "port", 0 );

		if( ( pcOption = pcOpts->FindOption( 'P' ) ) != 0 && pcOption->has_value )
			my->m_pcConn->m_cPassword = pcOption->value.const_str();
		else
			my->m_pcConn->m_cPassword = cSettings.GetString( "password" );

		if( ( pcOption = pcOpts->FindOption( 'e' ) ) != 0 && pcOption->has_value )
			my->m_pcConn->m_cDefaultEncoding = pcOption->value.const_str();
		else
			my->m_pcConn->m_cDefaultEncoding = cSettings.GetString( "encoding" );

		// get hardwired default values, if nothing was set
		// if no nick was given, use the login name from the os environment
		if( my->m_pcConn->m_cNick.empty() )
		{
			my->m_pcConn->m_cNick = getenv( "USER" );
			// only morons are using an irc client with a root account :P
			// i do it as of writing this too, so don't take it serious, it's a joke! ;)
			if( my->m_pcConn->m_cNick == "root" ) my->m_pcConn->m_cNick = "moron";
		}

		// if no server was given, connect to freenode (#syllable waits for YOU to join ;) )
		if( my->m_pcConn->m_cServer.empty() )
			my->m_pcConn->m_cServer = "irc.freenode.net";

		// if no port was given, use default irc port
		if( my->m_pcConn->m_nPort == 0 )
			my->m_pcConn->m_nPort = 6667;

		// if no default encoding is set, use ISO-8859-15, as this is default where I do live
		if( my->m_pcConn->m_cDefaultEncoding.empty() )
			my->m_pcConn->m_cDefaultEncoding = "ISO-8859-15";
	}

	// open iconv handler, only for encodings different from utf-8
	if( my->m_pcConn->m_cDefaultEncoding != "UTF-8" )
	{
		my->m_pcConn->m_hIconvIn = iconv_open( "UTF-8//TRANSLIT", my->m_pcConn->m_cDefaultEncoding.c_str() );
		my->m_pcConn->m_hIconvOut = iconv_open( os::String( my->m_pcConn->m_cDefaultEncoding+"//TRANSLIT" ).c_str(), "UTF-8" );
	}

	// create the window
	my->m_pcMainWindow = new MainWindow();
	my->m_pcMainWindow->CenterInScreen();
	my->m_pcMainWindow->Show();
	my->m_pcMainWindow->MakeFocus();

	// initialize the Main Window Messenger
	my->m_cMessengerWindowMain = os::Messenger( my->m_pcMainWindow );

	// tell the mainwindow what to show as nick
	os::Message cMessage( M_WND_CHANGENICK );
	cMessage.AddString( "nick", my->m_pcConn->m_cNick );
	my->m_cMessengerWindowMain.SendMessage( &cMessage );

	// warn the user, if invalid encoding is set
	if( my->m_pcConn->m_hIconvIn == (iconv_t)-1 || my->m_pcConn->m_hIconvOut == (iconv_t)-1 )
	{
		os::Message cMessageWindow( M_WND_DISPLAY );
		cMessageWindow.AddString( "display", "!!! invalid encoding set! take a look at http://www.gnu.org/software/libiconv/documentation/libiconv/iconv_open.3.html to see supported encodings\n" );
		my->m_cMessengerWindowMain.SendMessage( &cMessageWindow );
	}

	// start the socket Thread
	my->m_pcConn->m_pcSocketThread = new SocketThread();
	my->m_pcConn->m_pcSocketThread->Run();
	
	// initialize the Socket Messenger
	my->m_pcConn->m_cMessengerSocket = os::Messenger( my->m_pcConn->m_pcSocketThread );

	// connect the socket to a server
	cMessage = os::Message( M_SOCKET_CONNECT );
	cMessage.AddString( "server", my->m_pcConn->m_cServer );
	cMessage.AddInt16( "port", my->m_pcConn->m_nPort );
	my->m_pcConn->m_cMessengerSocket.SendMessage( &cMessage );
}

void App::HandleMessage( os::Message* pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		case M_SOCKET_CONNECTED:
		{
			os::Message cMessage( M_SOCKET_SEND );

			if( !my->m_pcConn->m_cPassword.empty() )
				cMessage.AddString( "send", os::String( "PASS " ) + my->OutputEncoding( my->m_pcConn->m_cPassword ) );

			cMessage.AddString( "send", os::String( "NICK " ) + my->OutputEncoding( my->m_pcConn->m_cNick ) );
			cMessage.AddString(
				"send",
				os::String().Format(
					"USER %s localhost %s :%s",
					my->OutputEncoding( my->m_pcConn->m_cNick ).c_str(),
					my->OutputEncoding( my->m_pcConn->m_cServer ).c_str(),
					my->OutputEncoding( my->m_pcConn->m_cNick ).c_str()
				)
			);
			my->m_pcConn->m_cMessengerSocket.SendMessage( &cMessage );
		}
		break;

		case M_APP_SEND:
		{
			os::String cString = "";
			os::String cSocketString = "";
			os::String cDisplayString = "";
			pcMessage->FindString( "send", &cString );
			
			// // means direct to socket
			if( cString.substr( 0, 2 ) == "//" )
			{
				cSocketString = cString.substr( 2, cString.Length() );
			}
			else if( cString.substr( 0, 1 ) == "/" )
			{
				// /? /h /help = help
				if( cString == "/?" || cString == "/h" || cString == "/help" )
				{
					cDisplayString = "*** help:\n";
					cDisplayString += "*** /? /h /help: this help message\n";
					cDisplayString += "*** /c: clear the window\n";
					cDisplayString += "*** /j #cannel: join a channel\n";
					cDisplayString += "*** This will automatically set the joined channel active.\n";
					cDisplayString += "*** /p: leave the active channel\n";
					cDisplayString += "*** /s #channel: set the active channel to this one\n";
					cDisplayString += "*** All messages you send will go to the active channel.\n";
					cDisplayString += "*** The active channel is shown in the bottom left next to your nick name.\n";
					cDisplayString += "*** /ls: list all open channels\n";
					cDisplayString += "*** /q: quit the connection and the application\n";
					cDisplayString += "*** //[SOMETHING] to send [SOMETHING] directly to the server\n";
					cDisplayString += "*** Please read the irc protocol rfc to know what can be done with this.\n";
					cDisplayString += "*** *** means Information from goIRC\n";
					cDisplayString += "*** !!! means Error from goIRC\n";
					cDisplayString += "*** ?>> means an unsupportet message (most of them right now) from the server, shown as it comes from there\n";
					cDisplayString += "*** <<? means goIRC did send an unknown message to the server, e.g. through //\n";
					cDisplayString += "*** Thanks for reading and have fun with goIRC, but keep in mind, this is alpha software!\n";
					cDisplayString += "*** end of help";
				}
				// /c = clear
				else if( cString == "/c" )
				{
					os::Message cMessageWindow( M_WND_CLEAR );
					my->m_cMessengerWindowMain.SendMessage( &cMessageWindow );
				}
				// /j = join
				else if( cString.substr( 1, 2 ) == "j " )
				{
					cSocketString = os::String( "JOIN " ) + cString.substr( 3 );
				}
				// /p = part active channel
				else if( cString.substr( 1, 2 ) == "p" && cString.Length() == 2 )
				{
					cSocketString = os::String( "PART " ) + my->m_pcConn->m_cChannel;
				}
				// /s = set the active channel
				else if( cString.substr( 1, 2 ) == "s " )
				{
					os::String cChannel = cString.substr( 3 );
					
					if( my->m_pcConn->m_cChannels.count( cChannel.c_str() ) > 0 )
					{
						my->m_pcConn->m_cChannel = cChannel;

						os::Message cMessage( M_WND_CHANGECHANNEL );
						cMessage.AddString( "channel", cChannel );
						cMessage.AddString( "topic", my->m_pcConn->m_cChannels[cChannel.c_str()] );
						my->m_cMessengerWindowMain.SendMessage( &cMessage );
					}
					else
					{
						cDisplayString = os::String( "channel " ) + cChannel + " not open, please join first (/j #channel) or check spelling (/ls to list open channels)";
					}
				}
				// /ls = list channels
				else if( cString.substr( 1, 2 ) == "ls" )
				{
					cDisplayString = "*** open channels:\n";
					std::map<os::String,os::String>::iterator Iterator = my->m_pcConn->m_cChannels.begin();
					while( Iterator != my->m_pcConn->m_cChannels.end() )
					{
						// FIXME: topic
						cDisplayString += os::String( "*** " ) + Iterator->first + " : " + Iterator->second.c_str() + "\n";
						++Iterator;
					}
					cDisplayString += "*** end of open channels";
				}
				// /q = quit
				else if( cString.substr( 1, 1 ) == "q" && cString.Length() == 2 )
				{
					os::Messenger( this ).SendMessage( M_APP_QUIT );
				}
				else
				{
					cDisplayString = os::String( "!!! unknown command (try /h or /? or /help): " ) + cString;
				}
			}
			else
			{
				if( !my->m_pcConn->m_cChannel.empty() )
				{
					cSocketString = os::String( "PRIVMSG " ) + my->m_pcConn->m_cChannel + " :" + cString;
				}
				else
				{
					cDisplayString = os::String( "!!! no channel set, nothing done! (" ) + cString + ")";
				}
			}

			if( !cSocketString.empty() )
			{
				os::Message cMessage = os::Message( M_SOCKET_SEND );
				cMessage.AddString( "send", my->OutputEncoding( cSocketString ) );
				my->m_pcConn->m_cMessengerSocket.SendMessage( &cMessage );
			}

			if( !cDisplayString.empty() )
			{
				os::Message cMessageWindow( M_WND_DISPLAY );
				cMessageWindow.AddString( "display", cDisplayString + "\n" );
				my->m_cMessengerWindowMain.SendMessage( &cMessageWindow );
			}
		}
		break;

		case M_SOCKET_ERROR:
		case M_SOCKET_SENT:
		case M_SOCKET_RECEIVED:
		case M_SOCKET_INFO:
		{
			os::String cString = "";
			os::String cOutput = "";

			switch( pcMessage->GetCode() )
			{				
				case M_SOCKET_ERROR:
				{
					int8 nSocketError;
					pcMessage->FindInt8( "errornumber", &nSocketError );
					cOutput = os::String().Format( "!!! Socket Error #%d\n", nSocketError );
				}
				break;

				case M_SOCKET_INFO:
				{
					pcMessage->FindString( "info", &cString );
					cOutput = os::String().Format( "*** %s\n", cString.c_str() );
				}
				break;

				case M_SOCKET_SENT:
				{
					pcMessage->FindString( "sent", &cString );
					IrcMessage cIrcMessage = my->m_pcIrcParser->ParseLine( cString );
					// don't show pong; join, part (the received answer will be shown)
					if(
						cIrcMessage.m_cCommand == "PONG" ||
						cIrcMessage.m_cCommand == "JOIN" ||
						cIrcMessage.m_cCommand == "PART"
					) {}
					// parse messages to produce readable output
					else if( cIrcMessage.m_cCommand == "PRIVMSG" )
					{
						cOutput =
							cIrcMessage.m_cVectorParams.at(0) + ", " +
							my->m_pcConn->m_cNick + ": " +
							my->InputEncoding( cIrcMessage.m_cVectorParams.at(1) ) + "\n";
					}
					// unknown message, display only
					else
					{
						cOutput = os::String( "<<? " ) + cString + "\n";
					}
				}
				break;

				case M_SOCKET_RECEIVED:
				{
					pcMessage->FindString( "received", &cString);
					cString = my->InputEncoding( cString );
					IrcMessage cIrcMessage = my->m_pcIrcParser->ParseLine( cString );

					// reply to ping with pong
					if( cIrcMessage.m_cCommand == "PING" )
					{
						os::Message cMessage( M_SOCKET_SEND );
						if( !cIrcMessage.m_cVectorParams.empty() )
							cMessage.AddString( "send", os::String( "PONG " ) + cIrcMessage.m_cVectorParams.at(0) );
						else
							cMessage.AddString( "send", os::String( "PONG" ) );
						my->m_pcConn->m_cMessengerSocket.SendMessage( &cMessage );
					}
					// parse messages to produce readable output
					else if( cIrcMessage.m_cCommand == "PRIVMSG" )
					{
						cOutput =
							cIrcMessage.m_cVectorParams.at(0) + " " +
							cIrcMessage.m_cSender + ": " +
							cIrcMessage.m_cVectorParams.at(1) + "\n";
					}
					// set channel on join
					else if( cIrcMessage.m_cCommand == "JOIN" )
					{
						if( cIrcMessage.m_cSender == my->m_pcConn->m_cNick )
						{
							my->m_pcConn->m_cChannel = cIrcMessage.m_cVectorParams.at(0);

							my->m_pcConn->m_cChannels[my->m_pcConn->m_cChannel] = "";
							
							os::Message cMessage( M_WND_CHANGECHANNEL );
							cMessage.AddString( "channel", my->m_pcConn->m_cChannel );
							cMessage.AddString( "topic", "" );
							my->m_cMessengerWindowMain.SendMessage( &cMessage );
	
							cOutput = os::String().Format( "*** joined %s\n", cIrcMessage.m_cVectorParams.at(0).c_str() );
						}
						else
						{
							cOutput =
								os::String( "*** " ) +
								cIrcMessage.m_cSender + "!" + cIrcMessage.m_cUser + "@" + cIrcMessage.m_cHost +
								" joined " + cIrcMessage.m_cVectorParams.at(0) + "\n";
						}
					}
					// remove channel on part
					else if( cIrcMessage.m_cCommand == "PART" )
					{
						if( cIrcMessage.m_cSender == my->m_pcConn->m_cNick )
						{
							my->m_pcConn->m_cChannels.erase( cIrcMessage.m_cVectorParams.at(0).c_str() );
							my->m_pcConn->m_cChannel = "";
	
							os::Message cMessage( M_WND_CHANGECHANNEL );
							cMessage.AddString( "channel", "---" );
							cMessage.AddString( "topic", "---" );
							my->m_cMessengerWindowMain.SendMessage( &cMessage );
							
							cOutput = os::String( "*** left " ) + cIrcMessage.m_cVectorParams.at(0) + "\n";
						}
						else
						{
							cOutput = os::String( "*** " ) + cIrcMessage.m_cSender + " left " + cIrcMessage.m_cVectorParams.at(0) + "\n";
						}
					}
					// topic replies
					else if(
						cIrcMessage.m_cCommand == "331" || cIrcMessage.m_cCommand == "RRL_NOTOPIC" ||
						cIrcMessage.m_cCommand == "332" || cIrcMessage.m_cCommand == "RPL_TOPIC" ||
						cIrcMessage.m_cCommand == "TOPIC"
					)
					{
						unsigned int nParams = cIrcMessage.m_cVectorParams.size();
						os::String cChannel = cIrcMessage.m_cVectorParams.at(nParams-2);
						os::String cTopic = cIrcMessage.m_cVectorParams.at(nParams-1);
						my->m_pcConn->m_cChannels[cChannel] = cTopic;
						
						if( cChannel == my->m_pcConn->m_cChannel )
						{
							os::Message cMessage( M_WND_CHANGETOPIC );
							cMessage.AddString( "topic", cTopic );
							my->m_cMessengerWindowMain.SendMessage( &cMessage );
						}
					}
					// unknown message, display only
					else
					{
						cOutput = os::String().Format( "?>> %s\n", cString.c_str() );
					}
				}
				break;
			}

			os::Message cMessageWindow( M_WND_DISPLAY );
			cMessageWindow.AddString( "display", cOutput.const_str() );
			my->m_cMessengerWindowMain.SendMessage( &cMessageWindow );
		}
		break;
		
		case M_APP_QUIT:
		{
			os::Message cMessage = os::Message( M_SOCKET_SEND );
			cMessage.AddString(
				"send",
				my->OutputEncoding( os::String( "QUIT :goIRC " ) + GOIRC_VERSION + " running on syllable. http://www.syllable.org" )
			);
			my->m_pcConn->m_cMessengerSocket.SendMessage( &cMessage );
			// FIXME: wait for closed socket and stuff
			sleep( 1 );
			my->m_pcConn->m_cMessengerSocket.SendMessage( os::M_QUIT );
			sleep( 1 );
			PostMessage( os::M_QUIT );
		}
		break;
		
		default:
		{
			std::cout << "not handled Message in Application: " << pcMessage->GetCode() << std::endl;
			Handler::HandleMessage( pcMessage );
		}
		break;
	}
}

App::~App()
{
	delete( my->m_pcIrcParser );
	iconv_close( my->m_pcConn->m_hIconvIn );
	iconv_close( my->m_pcConn->m_hIconvOut );
	delete( my );
}
